;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "julia-ts-mode" "20260616.1624"
  "Major mode for Julia source code using tree-sitter."
  '((emacs      "29.1")
    (julia-mode "0.4"))
  :url "https://github.com/ronisbr/julia-ts-mode"
  :commit "d53fb5b2c7e83223dcd8c7ae6bd5e1abf18665f6"
  :revdesc "d53fb5b2c7e8"
  :keywords '("julia" "languages" "tree-sitter"))
